"""Abstract bracket classes."""

import abc
import bisect
import statistics
import typing

from mutwo.events import basic
from mutwo.utilities import constants

__all__ = (
    "Bracket",
    "BracketContainer",
)

AbstractTime = typing.Any
AbstractTimeOrAbstractTimeRange = typing.Union[
    AbstractTime, typing.Tuple[AbstractTime, AbstractTime]
]


class Bracket(basic.SimultaneousEvent):
    def __init__(
        self,
        tagged_events: typing.Sequence[
            typing.Union[
                basic.TaggedSimpleEvent,
                basic.TaggedSequentialEvent,
                basic.TaggedSimultaneousEvent,
            ]
        ],
        start_or_start_range: AbstractTimeOrAbstractTimeRange,
        end_or_end_range: AbstractTimeOrAbstractTimeRange,
    ):
        super().__init__(tagged_events)
        self.start_or_start_range = start_or_start_range
        self.end_or_end_range = end_or_end_range

    @staticmethod
    def _get_mean_of_value_or_value_range(
        value_or_value_range: typing.Union[
            constants.Real, typing.Tuple[constants.Real, constants.Real]
        ],
    ) -> constants.Real:
        if isinstance(value_or_value_range, typing.Sequence):
            return statistics.mean(value_or_value_range)
        else:
            return value_or_value_range

    @property
    def start_or_start_range(self) -> AbstractTimeOrAbstractTimeRange:
        return self._start_or_start_range

    @start_or_start_range.setter
    def start_or_start_range(
        self, start_or_start_range: AbstractTimeOrAbstractTimeRange
    ):
        self._start_or_start_range = start_or_start_range

    @property
    def end_or_end_range(self) -> AbstractTimeOrAbstractTimeRange:
        return self._end_or_end_range

    @end_or_end_range.setter
    def end_or_end_range(self, end_or_end_range: AbstractTimeOrAbstractTimeRange):
        self._end_or_end_range = end_or_end_range

    @property
    @abc.abstractmethod
    def mean_start(self) -> constants.Real:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def mean_end(self) -> constants.Real:
        raise NotImplementedError


T = typing.TypeVar("T", bound=Bracket)


class BracketContainer(typing.Generic[T]):
    def __init__(self, brackets: typing.Sequence[T]):
        self._brackets = basic.SequentialEvent(
            sorted(brackets, key=lambda bracket: bracket.mean_start)
        )

    def __repr__(self) -> str:
        return "BracketContainer({})".format(tuple(self._brackets))

    def __iter__(self):
        return iter(self._brackets)

    def register(self, bracket: T):
        """Add new bracket to BracketContainer.

        :arg bracket: The bracket which shall be added.
        """

        start_of_brackets = tuple(bracket.mean_start for bracket in self._brackets)
        index = bisect.bisect_left(start_of_brackets, bracket.mean_start)
        self._brackets.insert(index, bracket)

    def filter(self, tag: str) -> typing.Tuple[T, ...]:
        """Filter and return all brackets which contain events with the requested tag

        :arg tag: The tag which shall be investigated.
        """

        filtered_brackets = []
        for bracket in self._brackets:
            for tagged_event in bracket:
                if tagged_event.tag == tag:
                    filtered_brackets.append(bracket)
                    break

        return tuple(filtered_brackets)



"""Module for mutwo specific exceptions."""

from mutwo import parameters

__all__ = (
    "ValueNotAssignedError",
    "OverlappingTimeBracketsError",
    "SplitUnavailableChildError",
)


class ValueNotAssignedError(Exception):
    def __init__(self):
        super().__init__(
            "TimeBracket values haven't been assigned yet, run"
            " 'assign_concrete_times' method first"
        )


class OverlappingTimeBracketsError(Exception):
    def __init__(self):
        super().__init__("Found overlapping time brackets!")


class SplitUnavailableChildError(Exception):
    def __init__(self, absolute_time: parameters.abc.DurationType):
        super().__init__(
            f"Can't split child at absolute time '{absolute_time}'. There is no child"
            " event available at the requested time."
        )



"""Module for Cage-like time brackets.

Time brackets are events which are defined by unspecified start and end times.
"""

import typing

from mutwo.events import basic
from mutwo.events import brackets
from mutwo.utilities import constants
from mutwo.utilities import exceptions

__all__ = (
    "TimeBracket",
    "TimeBracketContainer",
)


TimeOrTimeRange = typing.Union[
    constants.Real, typing.Tuple[constants.Real, constants.Real]
]


class TimeBracket(brackets.Bracket):
    def __init__(
        self,
        tagged_events: typing.Sequence[
            typing.Union[
                basic.TaggedSimpleEvent,
                basic.TaggedSequentialEvent,
                basic.TaggedSimultaneousEvent,
            ]
        ],
        start_or_start_range: TimeOrTimeRange,
        end_or_end_range: TimeOrTimeRange,
        seed: typing.Optional[int] = None,
    ):
        super().__init__(tagged_events, start_or_start_range, end_or_end_range)

        import random as time_bracket_random

        if seed:
            time_bracket_random.seed(seed)
        self._time_bracket_random = time_bracket_random

    def _assign_concrete_time(
        self, time_or_time_range: TimeOrTimeRange
    ) -> constants.Real:
        if hasattr(time_or_time_range, "__getitem__"):
            return self._time_bracket_random.uniform(*time_or_time_range)
        else:
            return time_or_time_range

    @brackets.Bracket.start_or_start_range.setter
    def start_or_start_range(self, start_or_start_range: TimeOrTimeRange):
        self._start_or_start_range = start_or_start_range
        if hasattr(self, "_assigned_start_time"):
            self.assign_concrete_times()

    @brackets.Bracket.end_or_end_range.setter
    def end_or_end_range(self, end_or_end_range: TimeOrTimeRange):
        self._end_or_end_range = end_or_end_range
        if hasattr(self, "_assigned_end_time"):
            self.assign_concrete_times()

    @property
    def mean_start(self) -> constants.Real:
        return brackets.Bracket._get_mean_of_value_or_value_range(
            self.start_or_start_range
        )

    @property
    def mean_end(self) -> constants.Real:
        return TimeBracket._get_mean_of_value_or_value_range(self.end_or_end_range)

    @property
    def duration(self) -> constants.Real:
        return self.mean_end - self.mean_start

    @property
    def assigned_start_time(self) -> constants.Real:
        try:
            return self._assigned_start_time
        except AttributeError:
            raise exceptions.ValueNotAssignedError()

    @property
    def assigned_end_time(self) -> constants.Real:
        try:
            return self._assigned_end_time
        except AttributeError:
            raise exceptions.ValueNotAssignedError()

    def assign_concrete_times(self):
        """Assign concrete values for start and end time"""

        self._assigned_start_time = self._assign_concrete_time(
            self.start_or_start_range
        )
        self._assigned_end_time = self._assign_concrete_time(self.end_or_end_range)


class TimeBracketContainer(brackets.BracketContainer[TimeBracket]):
    pass



"""Algorithms which are related to German-Dutch composer G. M. Koenig."""

import typing

import expenvelope


class Tendency(object):
    """Tendency offers an interface for dynamically changing minima / maxima areas.

    :param minima_curve: The curve which describes the smallest allowed value over the
        time axis.
    :type minima_curve: expenvelope.Envelope
    :param maxima_curve: The curve which describes the biggest allowed value over the
        time axis.
    :type maxima_curve: expenvelope.Envelope
    :param random_seed: The random seed which shall be set.
    :type random_seed: int

    The class is based on Gottfried Michael Koenigs algorithm of "Tendenz-Masken" in
    his program "Projekt 2" where those minima / maxima areas represent probability
    fields.

    **Example:**

    >>> import expenvelope
    >>> from mutwo.generators import koenig
    >>> minima_curve = expenvelope.Envelope.from_points((0, 0), (1, 1), (2, 0))
    >>> maxima_curve = expenvelope.Envelope.from_points((0, 1), (1, 2), (2, 3))
    >>> my_tendency = koenig.Tendency(minima_curve, maxima_curve)
    >>> my_tendency.value_at(0.5)
    0.6456692551041303
    >>> my_tendency.value_at(0.5)
    0.9549270045140213
    """

    def __init__(
        self,
        minima_curve: expenvelope.Envelope,
        maxima_curve: expenvelope.Envelope,
        random_seed: int = 100,
    ):
        import random

        random.seed(random_seed)

        self._assert_curves_are_valid(minima_curve, maxima_curve)
        self._minima_curve = minima_curve
        self._maxima_curve = maxima_curve
        self._random = random

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.minima_curve}, {self.maxima_curve})"

    @staticmethod
    def _assert_curves_are_valid(
        minima_curve: expenvelope.Envelope, maxima_curve: expenvelope.Envelope
    ):
        """Helper method that asserts the curves are a valid min/max pair."""

        # make sure both curves have equal duration
        try:
            assert minima_curve.end_time() == maxima_curve.end_time()
        except AssertionError:
            message = (
                "Found unequal duration when comparing 'minima_curve' (with end_time ="
                " '{}') and 'maxima_curve' (with end_time = '{}').".format(
                    minima_curve.end_time(), maxima_curve.end_time()
                )
            )
            message += " Make sure both curves have equal duration."
            raise ValueError(message)

        # compare all local extrema to make sure that at any time
        # point minima_curve < maxima_curve
        points_to_compare = (
            minima_curve.local_extrema(include_saddle_points=True)
            + maxima_curve.local_extrema(include_saddle_points=True)
            + [0, minima_curve.end_time()]
        )
        for time in points_to_compare:
            try:
                assert minima_curve.value_at(time) < maxima_curve.value_at(time)
            except AssertionError:
                message = (
                    "At time '{}' 'minima_curve' isn't smaller than 'maxima_curve'!"
                    .format(time)
                )
                raise ValueError(message)

    @property
    def minima_curve(self) -> expenvelope.Envelope:
        return self._minima_curve

    @minima_curve.setter
    def minima_curve(self, minima_curve: expenvelope.Envelope) -> expenvelope.Envelope:
        self._assert_curves_are_valid(minima_curve, self.maxima_curve)
        self._minima_curve = minima_curve

    @property
    def maxima_curve(self) -> expenvelope.Envelope:
        return self._maxima_curve

    @maxima_curve.setter
    def maxima_curve(self, maxima_curve: expenvelope.Envelope) -> expenvelope.Envelope:
        self._assert_curves_are_valid(self.minima_curve, maxima_curve)
        self._maxima_curve = maxima_curve

    def range_at(self, time: float) -> typing.Tuple[float]:
        """Get minima / maxima range at requested time."""
        return (self.minima_curve.value_at(time), self.maxima_curve.value_at(time))

    def value_at(self, time: float) -> float:
        """Get value at requested time."""
        return self._random.uniform(*self.range_at(time))



"""Routines to convert time brackets to basic events."""

import typing

from mutwo.converters import abc as converters_abc
from mutwo.events import basic
from mutwo.events import time_brackets
from mutwo.utilities import constants
from mutwo.utilities import exceptions

__all__ = ("TimeBracketsToEventConverter",)


class TimeBracketsToEventConverter(converters_abc.Converter):
    """Convert time bracket to basic events.

    :param tag: Events with this tag will be extracted and converted.
    """

    def __init__(self, tag: str):
        self._tag = tag

    @staticmethod
    def _get_start_and_end_time(
        time_bracket: time_brackets.TimeBracket,
    ) -> typing.Tuple[constants.Real, constants.Real]:
        try:
            return (
                time_bracket.assigned_start_time,
                time_bracket.assigned_end_time,
            )
        except exceptions.ValueNotAssignedError:
            time_bracket.assign_concrete_times()
            return TimeBracketsToEventConverter._get_start_and_end_time(time_bracket)

    def _extract_tagged_event(
        self,
        events: typing.Sequence[
            typing.Union[
                basic.TaggedSimpleEvent,
                basic.TaggedSequentialEvent,
                basic.TaggedSimultaneousEvent,
            ]
        ],
    ) -> typing.Union[
        basic.TaggedSimpleEvent,
        basic.TaggedSequentialEvent,
        basic.TaggedSimultaneousEvent,
    ]:
        for tagged_event in events:
            if tagged_event.tag == self._tag:
                return tagged_event

        raise NotImplementedError(f"Couldn't find instrument '{self._tag}' in bracket.")

    def convert(
        self, time_brackets_to_convert: typing.Sequence[time_brackets.TimeBracket]
    ) -> typing.Tuple[
        typing.Union[
            basic.TaggedSimpleEvent,
            basic.TaggedSequentialEvent,
            basic.TaggedSimultaneousEvent,
        ],
        ...,
    ]:
        events = []
        previous_end_time = 0
        for time_bracket in time_brackets_to_convert:
            (
                start_time,
                end_time,
            ) = TimeBracketsToEventConverter._get_start_and_end_time(time_bracket)

            # add rest if there is time between last time bracket and current bracket
            if previous_end_time < start_time:
                events.append(
                    basic.TaggedSimpleEvent(
                        start_time - previous_end_time, tag=self._tag
                    )
                )
            # raise error if last bracket is still running while the current
            # time bracket already started
            elif previous_end_time > start_time:
                raise exceptions.OverlappingTimeBracketsError()

            duration_in_seconds = end_time - start_time
            tagged_event = self._extract_tagged_event(
                tuple(time_bracket)
            ).destructive_copy()
            tagged_event.duration = duration_in_seconds
            events.append(tagged_event)

            previous_end_time = end_time

        return tuple(events)



"""Algorithms which are related to US-American theorist Erv Wilson."""

import functools
import operator
import itertools
import typing

from mutwo.parameters import pitches


def make_common_product_set_scale(
    numbers: typing.Sequence[int],
    n_combinations: int,
    tonality: bool,
    normalize: bool = False,
) -> typing.Tuple[pitches.JustIntonationPitch, ...]:
    """Make common product set scale as described in Wilsons letter to Fokker.

    :param numbers: The numbers which will be combined to single pitches.
    :type numbers: typing.Sequence[int]
    :param n_combinations: How many numbers will be combined for each pitch.
    :type n_combinations: int
    :param tonality: ``True`` for otonality and ``False`` for utonality.
    :type tonality: bool
    :param normalize: ``True`` if pitches shall become normalized to the same octave.
    :type normalize: bool

    **Example:**

    >>> from mutwo.generators import wilson
    >>> wilson.make_common_product_set_scale((3, 5, 7, 9), 2, True)
    (JustIntonationPitch(15),
     JustIntonationPitch(21),
     JustIntonationPitch(27),
     JustIntonationPitch(35),
     JustIntonationPitch(45),
     JustIntonationPitch(63))
    >>> wilson.make_common_product_set_scale((3, 5, 7, 9), 2, False)
    (JustIntonationPitch(1/15),
     JustIntonationPitch(1/21),
     JustIntonationPitch(1/27),
     JustIntonationPitch(1/35),
     JustIntonationPitch(1/45),
     JustIntonationPitch(1/63))
    """

    common_product_set_scale = []
    for combined_numbers in itertools.combinations(numbers, n_combinations):
        product = functools.reduce(operator.mul, combined_numbers)
        if tonality:
            ratio = f"{product}/1"
        else:
            ratio = f"1/{product}"

        common_product_set_scale.append(pitches.JustIntonationPitch(ratio))

    if normalize:
        for pitch in common_product_set_scale:
            pitch.normalize()

    return tuple(common_product_set_scale)



"""Constants for :mod:`mutwo.converters.backends.mmml`."""

EVENT_IDENTIFIER = " "
"""Identifier for a new event."""

OCTAVE_IDENTIFIER = ":"
"""Identifier for indicating the octave."""

MULTIPLE_PITCHES_IDENTIFIER = ","
"""Identifier for multiple pitches (for writing chords)."""

RHYTHM_IDENTIFIER = "'"
"""Identifier for writing rhythms."""

DYNAMIC_IDENTIFIER = '"'
"""Identifier for writing dynamics."""

ATTRIBUTE_IDENTIFIER = r"/"
"""Identifier for setting attributes.
Inspired by Lilypond and Guido."""

ATTRIBUTE_ARGUMENT_START_IDENTIFIER = "<"
"""Identifier for setting arguments of attributes.
Inspired by Guido."""

ATTRIBUTE_ARGUMENT_END_IDENTIFIER = ">"
"""Identifier for setting arguments of attributes.
Inspired by Guido."""

VARIABLE_IDENTIFIER = "$"

REST_IDENTIFIER = "r"
"""Identifier for writing rests.
Inspired by Lilypond."""

COMMENT_IDENTIFIER = "#"
"""Identifier for writing comments.
Inspired by bash and scripting languages like Python, Perl, etc."""

JUST_INTONATION_POSITIVE_EXPONENT_IDENTIFIER = "+"
"""Identifier for positive exponent when using
:class:`mutwo.converters.backends.mmml.MMMLSingleJIPitchConverter`."""


JUST_INTONATION_NEGATIVE_EXPONENT_IDENTIFIER = "-"
"""Identifier for negative exponent when using
:class:`mutwo.converters.backends.mmml.MMMLSingleJIPitchConverter`."""
